import random

class HashLinear:
    """Implementa uma tabela de Hash com encadeamento Linear"""
    _p = 2147483647

    def __init__(self):
        self._e = [None]*8
        self._a = random.randrange(HashLinear._p)
        self._b = random.randrange(HashLinear._p)
        self._n = 0

    def _getindex(self, key):
        return ((self._a*hash(key)+self._b)%HashLinear._p)%len(self._e)

    def _findpos(self, key):
        """Encontra a entrada na sequência que contém a chave key, ou uma entrada com None"""
        i = self._getindex(key)
        while self._e[i] and self._e[i][0]!=key:
            i = (i+1)%len(self._e)
        return i

    def __contains__(self, key):
        return bool(self._e[self._findpos(key)])

    def __setitem__(self, key, value):
        i = self._findpos(key)
        if self._e[i]:
            self._e[i] = key, value                        
        else: # Novo item
           self._e[i] = key, value
           self._n += 1
           self._checkcount()

    def __getitem__(self, key):
        i = self._findpos(key)
        if self._e[i] == None:
            raise KeyError
        else:
           return self._e[i][1]

    def __iterkeys__(self):
        for e in self._e:
            if e:
                yield e[0]

    def keys(self):
        return self.__iterkeys__()
    
    def __iter__(self):
        return self.keys()

    def __delitem__(self, key):
        i = self._findpos(key)
        if self._e[i] == None:
            raise KeyError        
        self._n -= 1
        j=i
        while True:
            self._e[i] = None
            skip = True;
            while skip:
                j = (j+1)%len(self._e)            
                if self._e[j] == None:
                    return 
                k = self._getindex(self._e[j][0])
                if i<j:
                   skip = (k>i) and (k<=j)
                else:
                   skip = (i<k) or (k>=j)               
            self._e[i] = self._e[j]
            i = j

    def __len__(self):
        return self._n
        
    def _checkcount(self):
        """ Limita o número de entradas a 75% da tabela"""
        if self._n*3 > len(self._e)*2:
            old_e = self._e
            self._e = [None]*(2*len(old_e))
            self._n = 0
            for e in old_e:                
                if e:
                    self[e[0]] = e[1]

class HashAberto:
    """Implementa uma tabela de hash com encadeamento em aberto"""
    _p = 2147483647
    
    class NoLL:
        """Nó da lista ligada que implementa os buckets"""
        def __init__(self, chave, valor, prox):
            self.chave = chave
            self.valor = valor
            self.prox = prox
    
    def __init__(self):
        self._e = [None]*8
        self._a = random.randrange(self._p)
        self._b = random.randrange(self._p)
        self._n = 0
    
    def _calchash(self, chave):
        return ((chave*self._a + self._b)%self._p)%len(self._e)
    
    def __setitem__(self, chave, valor):
        i = self._calchash(chave)
        # Procura dentro do vetor pelo elemento
        e = self._e[i]
        while e:
            if e.chave == chave: # Achei a chave, modifico valor
                e.valor = valor
                return
            e = e.prox
        # Não achei, insere novo
        self._e[i] = HashAberto.NoLL(chave, valor, self._e[i])
        self._n += 1
        self._verificatabela()
        
    def __getitem__(self, chave):
        i = self._calchash(chave)
        # Procura dentro do vetor pelo elemento
        e = self._e[i]
        while e:
            if e.chave == chave: # Achei a chave, modifico valor
                return e.valor
            e = e.prox
        # Não achei
        raise KeyError(chave)
        
    def __len__(self):
        return self._n
    
    def _verificatabela(self):
        """Limita o número de entradas a 100% do tamanho da tabela"""
        if self._n == len(self._e):
            # duplica
            old = self._e
            self._e = [None]*(2*len(old))
            self._n = 0
            # Reinsere elementos
            for e in old:
                while e:
                    self[e.chave] = e.valor
                    e = e.prox  
    
    def __iterkeys__(self):
        for e in self._e:
            while e:
                yield e.chave
                e = e.prox
            
    def __iter__(self):
        return self.__iterkeys__()
    
    def __delitem__(self, chave):
        i = self._calchash(chave)
        # Procura dentro do vetor pelo elemento
        e = self._e[i]
        if e:
            if e.chave == chave:
                # Caso especial: primeiro elemento
                self._e[i] = e.prox
                self._n -= 1
                return
            else:
                pai = e
                e = e.prox
                while e:
                    if e.chave == chave:
                        pai.prox = e.prox
                        self._n -= 1
                        return
                    pai = e
                    e = e.prox
        raise KeyError(chave)
                    